package dk.au.daimi.rabbithunt.control;

import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.util.Iterator;
import java.util.Random;
import java.util.Vector;
import javax.swing.Icon;
import dk.au.daimi.rabbithunt.model.Animal;
import dk.au.daimi.rabbithunt.model.Bush;
import dk.au.daimi.rabbithunt.model.Fox;
import dk.au.daimi.rabbithunt.model.FoxBrain;
import dk.au.daimi.rabbithunt.model.Piece;
import dk.au.daimi.rabbithunt.model.Rabbit;
import dk.au.daimi.rabbithunt.model.RabbitBrain;
import dk.au.daimi.rabbithunt.model.world.Direction;
import dk.au.daimi.rabbithunt.model.world.World;
import dk.au.daimi.rabbithunt.view.Grid;

/**
 * Creates and maintains a "field" for the rabbit and fox, and sets up a number
 * of methods to be used to move animals around. The brains are not given direct
 * access to these methods, as this would allow them to cheat. Instead, they are
 * given limited access to these methods through a World object that represents
 * only the things they are allowed to know.
 * 
 * @author David Matuszek
 * @author Martin Geisler
 * @author Mikkel Krigrd
 * @version August 16, 2005
 */
public class Controller
{
	private static Bush BUSH_PIECE = new Bush();
	public long oldRandomSeed = 30;
	// define instance variables
	private Grid<Piece> field;
	private int numberOfRows;
	private int numberOfColumns;
	private Vector<AnimalInWorld> animals;
	private int rabbitIndex;
	private static Random randomNumberGenerator = new Random();
	private static Random seedGenerator = new Random();
	private boolean lookLines = false;
	/** True if rabbit's turn, false if fox's turn */
	private boolean isRabbitsTurn;
	/** Flag to tell when the game ends */
	private boolean gameIsOver;
	/** The number of turns taken so far in the game */
	private int stepsTaken;
	private int numberOfFoxes;
	private Rabbit rabbit;
	/**
	 * Constructs a model that uses a field of the given size.
	 * 
	 * @param width the width of the field to be used by the model
	 * @param height the height of the field to be used by the model
	 */
	public Controller(int width, int height)
	{
		this(width, height, 1);
	}
	public Controller(int width, int height, int number_of_foxes)
	{
		field = new Grid<Piece>(width, height);
		this.numberOfFoxes = number_of_foxes;
		numberOfRows = width;
		numberOfColumns = height;
		reset();
	}
	/**
	 * Returns the field.
	 * 
	 * @return the field
	 */
	public Grid<Piece> getField()
	{
		return field;
	}
	/**
	 * Sets replay mode and rewinds to turn 0. The game that starts there is
	 * exactly the same as before.
	 */
	public void replay()
	{
		populate(true);
	}
	/**
	 * Reset the game.
	 */
	public void reset()
	{
		populate(false);
	}
	/**
	 * Puts a rabbit, a fox, and some bushes in the field.
	 * 
	 * @param replay
	 */
	private void populate(boolean replay)
	{
		this.animals = new Vector<AnimalInWorld>();
		gameIsOver = false;
		stepsTaken = 0;
		// if this is not a replay, we use a new random seed
		if (!replay)
		{
			//oldRandomSeed =-7309212569308854764L; //Worst-Case
			oldRandomSeed = seedGenerator.nextLong();
		}
		randomNumberGenerator.setSeed(oldRandomSeed);
		// create some abbreviations, just to save some typing
		int numRows = numberOfRows;
		int numCols = numberOfColumns;
		// remove any previous contents of field
		for (int i = 0; i < numRows; i++)
			for (int j = 0; j < numCols; j++)
				field.set(i, j, null);
		// put the rabbit in a random location
		int rabbitRow = random(0, numRows - 1);
		int rabbitColumn = random(0, numCols - 1);
		rabbit = new Rabbit(this);
		rabbit.setBrain(new RabbitBrain());
		AnimalInWorld rab = new AnimalInWorld(rabbitRow, rabbitColumn, rabbit);
		this.animals.add(rab);
		field.set(rabbitRow, rabbitColumn, rabbit);
		this.rabbitIndex = this.animals.indexOf(rab);
		for (int i = 0; i < numberOfFoxes; i++)
		{
			// put the fox in a random location, not too close to the rabbit
			int distance;
			int foxRow;
			int foxColumn;
			do
			{
				foxRow = random(0, numRows - 1);
				foxColumn = random(0, numCols - 1);
				distance = Math.max(Math.abs(foxRow - rabbitRow), Math.abs(foxColumn
						- rabbitColumn));
			} while (distance < (numRows + numCols) / 4);
			Fox fox = new Fox(this);
			fox.setBrain(new FoxBrain());
			field.set(foxRow, foxColumn, fox);
			this.animals.add(new AnimalInWorld(foxRow, foxColumn, fox));
		}
		// put in some random bushes
		// (since bushes don't do anything, we cheat and use the same bush)
		int numberOfBushes = (numRows * numCols) / 20;
		for (int i = 0; i < numberOfBushes; i++)
		{
			int bushRow = random(0, numRows - 1);
			int bushColumn = random(0, numCols - 1);
			if (field.get(bushRow, bushColumn) == null)
			{
				field.set(bushRow, bushColumn, BUSH_PIECE);
			} else
				i--;
		}
	}
	/**
	 * Gives one animal a chance to move.
	 */
	public void allowSingleMove()
	{
		Animal animal;
		Direction direction;
		int row, column;
		Iterator<AnimalInWorld> iaiw = animals.iterator();
		while (iaiw.hasNext())
		{
			AnimalInWorld aiw = iaiw.next();
			// make sure it's legal to allow moves
			if (gameIsOver)
				return;
			animal = aiw.getAni();
			row = aiw.getRow();
			column = aiw.getCol();
			// ask the animal to decide a direction
			direction = animal.decideMove();
			// if move is legal, do it
			if (direction != Direction.NONE)
			{
				int newRow = row + rowChange(direction);
				int newColumn = column + columnChange(direction);
				if (legalLocation(newRow, newColumn)
						&& field.get(newRow, newColumn) != BUSH_PIECE)
				{
					// perform move
					field.set(row, column, null);
					field.set(newRow, newColumn, animal);
					aiw.setRow(newRow);
					aiw.setCol(newColumn);
				}
			}
			// check whether move was fatal for rabbit
			if (animals.get(this.rabbitIndex) != aiw
					&& animals.get(this.rabbitIndex).getRow() == aiw.getRow()
					&& animals.get(this.rabbitIndex).getCol() == aiw.getCol())
			{
				gameIsOver = true;
			}
		}
		stepsTaken++;
	}
	/**
	 * Returns the number of steps taken so far.
	 * 
	 * @return the number of steps taken
	 */
	public int stepsTaken()
	{
		return stepsTaken;
	}
	/**
	 * Is the game over?
	 * 
	 * @return true if the game is over, otherwise false
	 */
	public boolean gameIsOver()
	{
		return gameIsOver;
	}
	/**
	 * Utility method to choose a random integer from min to max, inclusive.
	 * 
	 * @param min the smallest number to be returned
	 * @param max the largest number to be returned
	 * @return a random number N, where min &lt;= N &lt;= max
	 */
	public static int random(int min, int max)
	{
		return randomNumberGenerator.nextInt(max - min + 1) + min;
	}
	/**
	 * Determines whether the given row and column numbers represent a legal
	 * location in the field.
	 * 
	 * @param row the row number
	 * @param column the column number
	 */
	public boolean legalLocation(int row, int column)
	{
		return row >= 0 && row < numberOfRows && column >= 0
				&& column < numberOfColumns;
	}
	/**
	 * Toggles look lines on the field. These are just lines of dots made when an
	 * animal looks in a certain direction. This means placing a "dot" piece in
	 * the field for while and is just for drawing purposes.
	 * 
	 * @param lookLines true if we want look lines, otherwise false
	 */
	public void setLookLines(boolean lookLines)
	{
		this.lookLines = lookLines;
	}
	/**
	 * Determines what can be seen from a given location, looking in a given
	 * direction.
	 * 
	 * @param animal the animal making the look
	 * @param direction the direction of the look
	 * @return the object seen
	 */
	public int look(Animal animal, Direction direction)
	{
		int row, column;
		int idx = -1;
		Iterator<AnimalInWorld> iain = animals.iterator();
		while (iain.hasNext())
		{
			AnimalInWorld aninwo = iain.next();
			if (aninwo.getAni() == animal)
			{
				idx = animals.indexOf(aninwo);
			}
		}
		if (idx != -1)
		{
			row = animals.get(idx).getRow();
			column = animals.get(idx).getCol();
		} else
		{
			throw new IllegalArgumentException(
					"Animal must be either the fox or the rabbit.");
		}
		// decode direction into its x-y components
		int rowDelta = rowChange(direction);
		int columnDelta = columnChange(direction);
		Piece look = new Look();
		int result = World.EDGE;
		// number of look-dots placed
		int n = 0;
		while (true)
		{
			row += rowDelta;
			column += columnDelta;
			if (!legalLocation(row, column))
			{
				result = World.EDGE;
				break;
			}
			// System.out.println("Looking with thread: " + Thread.currentThread());
			Piece p = field.get(row, column);
			if (p == null)
			{
				if (lookLines)
				{
					field.set(row, column, look);
					n++;
					try
					{
						Thread.sleep(25);
					} catch (InterruptedException e)
					{
						// Ignored.
					}
				}
			} else
			{
				if (p instanceof Rabbit)
					result = World.RABBIT;
				if (p instanceof Fox)
					result = World.FOX;
				if (p instanceof Bush)
					result = World.BUSH;
				break;
			}
		}
		if (lookLines)
		{
			try
			{
				Thread.sleep(100);
			} catch (InterruptedException e)
			{
				// Ignored.
			}
			while (n > 0)
			{
				row -= rowDelta;
				column -= columnDelta;
				n--;
				field.set(row, column, null);
			}
		}
		return result;
	}
	/**
	 * Determines how moving in the given direction affects the row number.
	 * 
	 * @param direction the direction in which to move
	 * @return the amount by which the row number will change
	 */
	private int rowChange(Direction direction)
	{
		if (direction == Direction.N || direction == Direction.NE
				|| direction == Direction.NW)
			return -1;
		if (direction == Direction.S || direction == Direction.SE
				|| direction == Direction.SW)
			return +1;
		return 0;
	}
	/**
	 * Determines how moving in the given direction affects the column number.
	 * 
	 * @param direction the direction in which to move
	 * @return the amount by which the column number will change
	 */
	private int columnChange(Direction direction)
	{
		if (direction == Direction.W || direction == Direction.NW
				|| direction == Direction.SW)
			return -1;
		if (direction == Direction.E || direction == Direction.NE
				|| direction == Direction.SE)
			return +1;
		return 0;
	}
	/**
	 * Determines the distance to the nearest thing, or to the edge of the field,
	 * looking in a given direction.
	 * 
	 * @param animal the animal measuring the distance.
	 * @param direction the direction of the look
	 * @return the distance
	 */
	public int distance(Animal animal, Direction direction)
	{
		int row, column;
		int idx = -1;
		Iterator<AnimalInWorld> iain = animals.iterator();
		while (iain.hasNext())
		{
			AnimalInWorld aninwo = iain.next();
			if (aninwo.getAni() == animal)
			{
				idx = animals.indexOf(aninwo);
			}
		}
		if (idx != -1)
		{
			row = animals.get(idx).getRow();
			column = animals.get(idx).getCol();
		} else
		{
			throw new IllegalArgumentException("Animal must be either the "
					+ "fox or the rabbit.");
		}
		// decode direction into its x-y components
		int rowDelta = rowChange(direction);
		int columnDelta = columnChange(direction);
		// check in that direction until you see something
		// (if nothing else, you will eventually see the edge of the
		// array, thus the loop <i>will</i> terminate)
		int steps = 0;
		while (true)
		{
			row = row + rowDelta;
			column = column + columnDelta;
			steps++;
			if (!legalLocation(row, column) || field.get(row, column) != null)
			{
				return steps;
			}
		}
	}
	/**
	 * Is it the rabbit's turn?
	 * 
	 * @return true if it is the rabbit's turn, false otherwise
	 */
	public boolean isRabbitsTurn()
	{
		return isRabbitsTurn;
	}

	public class AnimalInWorld
	{
		private int row;
		private int col;
		private Animal ani;

		public AnimalInWorld(int row, int col, Animal ani)
		{
			this.row = row;
			this.col = col;
			this.ani = ani;
		}
		public int getCol()
		{
			return col;
		}
		public void setCol(int col)
		{
			this.col = col;
		}
		public int getRow()
		{
			return row;
		}
		public void setRow(int row)
		{
			this.row = row;
		}
		public Animal getAni()
		{
			return ani;
		}
	}

	// a small class representing the dots that appear when an animal
	// looks in a certain direction
	private static class Look extends Piece
	{
		public Look()
		{
			setIcon(new LookIcon());
		}

		private static class LookIcon implements Icon
		{
			public void paintIcon(Component c, Graphics g, int x, int y)
			{
				if (g instanceof Graphics2D)
					((Graphics2D) g).setRenderingHint(RenderingHints.KEY_ANTIALIASING,
							RenderingHints.VALUE_ANTIALIAS_ON);
				g.setColor(Color.YELLOW);
				g.fillOval(x + SIZE / 4, y + SIZE / 4, x + SIZE / 2, y + SIZE / 2);
			}
			public int getIconWidth()
			{
				return Piece.SIZE;
			}
			public int getIconHeight()
			{
				return Piece.SIZE;
			}
		}
	}
}
