package dk.au.daimi.rabbithunt.model.world;
import dk.au.daimi.rabbithunt.control.Controller;
import dk.au.daimi.rabbithunt.model.Animal;

/**
 * This class represents the world seen through the eyes of a
 * particular animal. It contains some constants defining the objects
 * that can be seen in the world, and it provides methods for
 * navigation.
 *
 * The purpose of this class is to clearly define what an animal is
 * supposed to know about the world as opposed to what the program is
 * allowed to know. This way the brain does not use the model directly
 * but only an isolated part, making it harder to cheat.
 *
 * @author David Matuszek
 * @author Martin Geisler
 * @author Mikkel Krigrd
 * @version August 16, 2005
 */
public class World {

    // These constants define the things that can be seen in the world.
    // As we all know, the world is flat and has edges.
    public static final int EDGE   = 0;
    public static final int FOX    = 1;
    public static final int RABBIT = 2;
    public static final int BUSH   = 3;

    private Animal animal;
    private Controller model;

    /**
     * Creates a new world for an animal and a model.
     * @param a the animal
     * @param m the model
     */
    public World(Animal a, Controller m) {
        animal = a;
        model = m;
    }

    /**
     * Returns an integer indicating which object is the closest in
     * the given direction. This integer matches one of the constants
     * defined in this class.
     * 
     * @param direction the direction to look in
     * @return the closest object in the given direction
    */
    public int look(Direction direction) {
        return model.look(animal, direction);
    }

    /**
     * Determines whether this animal can move one step in the
     * given direction. An animal cannot move off the edge of
     * the array or onto a bush, but it <i>can</i> move onto
     * another animal.<p>
     * If we extend this program to allow multiple rabbits or
     * multiple foxes, we should disallow an animal moving onto
     * another animal of the same type.
     *
     * @param direction the direction in which to consider moving
     * @return true if the move is possible
     */
    public boolean canMove(Direction direction) {
        if (direction == Direction.NONE)
            return true;
        if (distance(direction) > 1)
            return true;
        int object = look(direction);
        if (object == EDGE || object == BUSH) return false;
        return true;
    }

    /**
     * Finds the distance to the first visible object in the given
     * direction, starting from this animal's current position.
     *
     * @param direction the direction in which to look
     * @return the distance to the object seen
     */
    public int distance(Direction direction) {
        return model.distance(animal, direction);
    }

    /**
     * Returns a random integer between lower bound
     * and upper bound, both inclusive.
     * @param lowerBound the lower bound
     * @param upperBound the upperBound
     * @return the random integer
     */
    public int random(int lowerBound, int upperBound) {
        return Controller.random(lowerBound, upperBound);
    }
}